@@ -33,12 +33,10 @@ is_known_tag({xmlel, Name, Attrs, _}, TopXMLNS) ->
     XMLNS = get_attr(<<"xmlns">>, Attrs, TopXMLNS),
     get_mod(Name, XMLNS) /= undefined.
 
-pp(Term) ->
-    case get_mod(Term) of
-      undefined ->
-	  io_lib_pretty:print(Term, fun (_, _) -> no end);
-      Mod -> io_lib_pretty:print(Term, fun Mod:pp/2)
-    end.
+get_els(Term) -> Mod = get_mod(Term), Mod:get_els(Term).
+
+set_els(Term, Els) ->
+    Mod = get_mod(Term), Mod:set_els(Term, Els).
 
 do_decode(Name, <<>>, _, _) ->
     erlang:error({xmpp_codec, {missing_tag_xmlns, Name}});
@@ -197,6 +195,15 @@ recompile_resolver(Mods, ResolverMod) ->
 					     "nofile", Code),
     ok.
 
+pp(xmlel, 3) -> [name, attrs, children];
+pp(Name, Arity) ->
+    case get_mod(erlang:make_tuple(Arity + 1, undefined,
+				   [{1, Name}]))
+	of
+      undefined -> no;
+      Mod -> Mod:pp(Name, Arity)
+    end.
+
 records() -> [].
 
 get_mod(<<"status">>, <<"jabber:server">>) -> rfc6120;
@@ -491,6 +498,9 @@ get_mod(<<"slot">>,
     xep0363;
 get_mod(<<"get">>, <<"urn:xmpp:http:upload:0">>) ->
     xep0363;
+get_mod(<<"pointer">>,
+	<<"urn:xmpp:avatar:metadata">>) ->
+    xep0084;
 get_mod(<<"not-authorized">>,
 	<<"urn:ietf:params:xml:ns:xmpp-stanzas">>) ->
     rfc6120;
@@ -557,6 +567,7 @@ get_mod(<<"system-shutdown">>,
 get_mod(<<"items">>,
 	<<"http://jabber.org/protocol/pubsub#event">>) ->
     xep0060;
+get_mod(<<"query">>, <<"urn:xmpp:mam:2">>) -> xep0313;
 get_mod(<<"prev">>,
 	<<"http://jabber.org/protocol/commands">>) ->
     xep0050;
@@ -579,6 +590,8 @@ get_mod(<<"redirect">>,
 	<<"http://jabber.org/protocol/pubsub#owner">>) ->
     xep0060;
 get_mod(<<"a">>, <<"urn:xmpp:sm:3">>) -> xep0198;
+get_mod(<<"info">>, <<"urn:xmpp:avatar:metadata">>) ->
+    xep0084;
 get_mod(<<"not-allowed">>,
 	<<"urn:ietf:params:xml:ns:xmpp-stanzas">>) ->
     rfc6120;
@@ -596,24 +609,14 @@ get_mod(<<"JABBERID">>, <<"vcard-temp">>) -> xep0054;
 get_mod(<<"subscription">>,
 	<<"http://jabber.org/protocol/pubsub#event">>) ->
     xep0060;
-get_mod(<<"iq">>, <<"jabber:iq:privacy">>) -> xep0016;
-get_mod(<<"list">>, <<"jabber:iq:privacy">>) -> xep0016;
 get_mod(<<"key">>, <<"jabber:iq:register">>) -> xep0077;
-get_mod(<<"CATEGORIES">>, <<"vcard-temp">>) -> xep0054;
-get_mod(<<"delete">>,
-	<<"http://jabber.org/protocol/pubsub">>) ->
-    xep0060;
 get_mod(<<"unsupported">>,
 	<<"http://jabber.org/protocol/pubsub#errors">>) ->
     xep0060;
 get_mod(<<"unsupported-access-model">>,
 	<<"http://jabber.org/protocol/pubsub#errors">>) ->
     xep0060;
-get_mod(<<"subscribe">>, <<"urn:xmpp:mix:0">>) ->
-    xep0369;
-get_mod(<<"nick">>, <<"jabber:iq:search">>) -> xep0055;
 get_mod(<<"ip">>, <<"urn:xmpp:sic:1">>) -> xep0279;
-get_mod(<<"port">>, <<"urn:xmpp:sic:1">>) -> xep0279;
 get_mod(<<"attribute">>, <<"urn:xmpp:delegation:1">>) ->
     xep0355;
 get_mod(<<"status">>, <<"jabber:component:accept">>) ->
@@ -621,28 +624,13 @@ get_mod(<<"status">>, <<"jabber:component:accept">>) ->
 get_mod(<<"query">>, <<"jabber:iq:register">>) ->
     xep0077;
 get_mod(<<"CLASS">>, <<"vcard-temp">>) -> xep0054;
-get_mod(<<"desc">>, <<"jabber:x:data">>) -> xep0004;
-get_mod(<<"no-permanent-store">>,
-	<<"urn:xmpp:hints">>) ->
-    xep0334;
-get_mod(<<"filename">>, <<"urn:xmpp:http:upload">>) ->
-    xep0363;
-get_mod(<<"storage">>, <<"storage:bookmarks">>) ->
-    xep0048;
-get_mod(<<"LINE">>, <<"vcard-temp">>) -> xep0054;
-get_mod(<<"PRODID">>, <<"vcard-temp">>) -> xep0054;
-get_mod(<<"publish-options">>,
-	<<"http://jabber.org/protocol/pubsub">>) ->
-    xep0060;
+get_mod(<<"result">>, <<"urn:xmpp:mam:2">>) -> xep0313;
 get_mod(<<"purge">>,
 	<<"http://jabber.org/protocol/pubsub#event">>) ->
     xep0060;
 get_mod(<<"header">>,
 	<<"http://jabber.org/protocol/shim">>) ->
     xep0131;
-get_mod(<<"reason">>,
-	<<"http://jabber.org/protocol/muc#admin">>) ->
-    xep0045;
 get_mod(<<"failed">>, <<"urn:xmpp:sm:3">>) -> xep0198;
 get_mod(<<"x">>, <<"jabber:x:expire">>) -> xep0023;
 get_mod(<<"policy-violation">>,
@@ -680,7 +668,11 @@ get_mod(<<"items">>,
 get_mod(<<"password">>,
 	<<"http://jabber.org/protocol/muc#user">>) ->
     xep0045;
+get_mod(<<"jid">>, <<"urn:xmpp:mam:2">>) -> xep0313;
 get_mod(<<"never">>, <<"urn:xmpp:mam:1">>) -> xep0313;
+get_mod(<<"metadata">>,
+	<<"urn:xmpp:avatar:metadata">>) ->
+    xep0084;
 get_mod(<<"bad-format">>,
 	<<"urn:ietf:params:xml:ns:xmpp-streams">>) ->
     rfc6120;
@@ -773,6 +765,8 @@ get_mod(<<"affiliation">>,
 get_mod(<<"affiliations">>,
 	<<"http://jabber.org/protocol/pubsub">>) ->
     xep0060;
+get_mod(<<"data">>, <<"urn:xmpp:avatar:data">>) ->
+    xep0084;
 get_mod(<<"item">>,
 	<<"http://jabber.org/protocol/disco#items">>) ->
     xep0030;
@@ -837,6 +831,8 @@ get_mod(<<"unsupported-feature">>,
 	<<"urn:ietf:params:xml:ns:xmpp-streams">>) ->
     rfc6120;
 get_mod(<<"DESC">>, <<"vcard-temp">>) -> xep0054;
+get_mod(<<"request">>, <<"urn:xmpp:receipts">>) ->
+    xep0184;
 get_mod(<<"url">>, <<"storage:bookmarks">>) -> xep0048;
 get_mod(<<"unexpected-request">>,
 	<<"urn:ietf:params:xml:ns:xmpp-stanzas">>) ->
@@ -950,6 +946,8 @@ get_mod(<<"URL">>, <<"vcard-temp">>) -> xep0054;
 get_mod(<<"headers">>,
 	<<"http://jabber.org/protocol/shim">>) ->
     xep0131;
+get_mod(<<"received">>, <<"urn:xmpp:receipts">>) ->
+    xep0184;
 get_mod(<<"jid-malformed">>,
 	<<"urn:ietf:params:xml:ns:xmpp-stanzas">>) ->
     rfc6120;
@@ -972,6 +970,7 @@ get_mod(<<"block">>, <<"urn:xmpp:blocking">>) ->
 get_mod(<<"delete">>,
 	<<"http://jabber.org/protocol/pubsub#event">>) ->
     xep0060;
+get_mod(<<"prefs">>, <<"urn:xmpp:mam:2">>) -> xep0313;
 get_mod(<<"sm">>, <<"urn:xmpp:sm:3">>) -> xep0198;
 get_mod(<<"fetch">>,
 	<<"http://jabber.org/protocol/offline">>) ->
@@ -1000,6 +999,9 @@ get_mod(<<"invalid-id">>,
 	<<"urn:ietf:params:xml:ns:xmpp-streams">>) ->
     rfc6120;
 get_mod(<<"PUBLIC">>, <<"vcard-temp">>) -> xep0054;
+get_mod(<<"precondition-not-met">>,
+	<<"http://jabber.org/protocol/pubsub#errors">>) ->
+    xep0060;
 get_mod(<<"item">>,
 	<<"http://jabber.org/protocol/offline">>) ->
     xep0013;
@@ -1088,9 +1090,11 @@ get_mod(<<"PAGER">>, <<"vcard-temp">>) -> xep0054;
 get_mod(<<"MSG">>, <<"vcard-temp">>) -> xep0054;
 get_mod(<<"PHONETIC">>, <<"vcard-temp">>) -> xep0054;
 get_mod(<<"jid">>, <<"urn:xmpp:mam:0">>) -> xep0313;
+get_mod(<<"fin">>, <<"urn:xmpp:mam:2">>) -> xep0313;
 get_mod(<<"ver">>, <<"urn:xmpp:features:rosterver">>) ->
     rfc6121;
 get_mod(<<"body">>, <<"jabber:client">>) -> rfc6120;
+get_mod(<<"always">>, <<"urn:xmpp:mam:2">>) -> xep0313;
 get_mod(<<"composing">>, <<"jabber:x:event">>) ->
     xep0022;
 get_mod(<<"reset">>,
@@ -1206,6 +1210,7 @@ get_mod(<<"PREF">>, <<"vcard-temp">>) -> xep0054;
 get_mod(<<"reason">>,
 	<<"http://jabber.org/protocol/muc#user">>) ->
     xep0045;
+get_mod(<<"never">>, <<"urn:xmpp:mam:2">>) -> xep0313;
 get_mod(<<"stat">>,
 	<<"http://jabber.org/protocol/stats">>) ->
     xep0039;
@@ -1307,6 +1312,32 @@ get_mod(<<"restricted-xml">>,
     rfc6120;
 get_mod(<<"HOME">>, <<"vcard-temp">>) -> xep0054;
 get_mod(<<"end">>, <<"urn:xmpp:mam:tmp">>) -> xep0313;
+get_mod(<<"iq">>, <<"jabber:iq:privacy">>) -> xep0016;
+get_mod(<<"list">>, <<"jabber:iq:privacy">>) -> xep0016;
+get_mod(<<"CATEGORIES">>, <<"vcard-temp">>) -> xep0054;
+get_mod(<<"delete">>,
+	<<"http://jabber.org/protocol/pubsub">>) ->
+    xep0060;
+get_mod(<<"subscribe">>, <<"urn:xmpp:mix:0">>) ->
+    xep0369;
+get_mod(<<"nick">>, <<"jabber:iq:search">>) -> xep0055;
+get_mod(<<"port">>, <<"urn:xmpp:sic:1">>) -> xep0279;
+get_mod(<<"desc">>, <<"jabber:x:data">>) -> xep0004;
+get_mod(<<"no-permanent-store">>,
+	<<"urn:xmpp:hints">>) ->
+    xep0334;
+get_mod(<<"filename">>, <<"urn:xmpp:http:upload">>) ->
+    xep0363;
+get_mod(<<"storage">>, <<"storage:bookmarks">>) ->
+    xep0048;
+get_mod(<<"LINE">>, <<"vcard-temp">>) -> xep0054;
+get_mod(<<"PRODID">>, <<"vcard-temp">>) -> xep0054;
+get_mod(<<"publish-options">>,
+	<<"http://jabber.org/protocol/pubsub">>) ->
+    xep0060;
+get_mod(<<"reason">>,
+	<<"http://jabber.org/protocol/muc#admin">>) ->
+    xep0045;
 get_mod(Name, XMLNS) ->
     xmpp_codec_external:lookup(Name, XMLNS).
 
@@ -1315,16 +1346,10 @@ get_mod({legacy_auth_feature}) -> xep0078;
 get_mod({carbons_disable}) -> xep0280;
 get_mod({last, _, _}) -> xep0012;
 get_mod({compression, _}) -> xep0138;
-get_mod({caps, _, _, _, _}) -> xep0115;
-get_mod({xdata, _, _, _, _, _, _}) -> xep0004;
-get_mod({ps_options, _, _, _, _}) -> xep0060;
-get_mod({muc_admin, _}) -> xep0045;
 get_mod({mam_prefs, _, _, _, _}) -> xep0313;
-get_mod({search, _, _, _, _, _, _, _}) -> xep0055;
-get_mod({expire, _, _}) -> xep0023;
-get_mod({privilege, _, _}) -> xep0356;
 get_mod({stanza_error, _, _, _, _, _, _}) -> rfc6120;
 get_mod({muc_user, _, _, _, _, _, _}) -> xep0045;
+get_mod({avatar_pointer, _, _, _, _, _, _}) -> xep0084;
 get_mod({bytestreams, _, _, _, _, _, _}) -> xep0065;
 get_mod({muc_invite, _, _, _, _}) -> xep0045;
 get_mod({sasl_auth, _, _}) -> rfc6120;
@@ -1401,6 +1426,7 @@ get_mod({muc, _, _}) -> xep0045;
 get_mod({carbons_enable}) -> xep0280;
 get_mod({carbons_private}) -> xep0280;
 get_mod({mix_leave}) -> xep0369;
+get_mod({avatar_info, _, _, _, _, _, _}) -> xep0084;
 get_mod({privacy_list, _, _}) -> xep0016;
 get_mod({text, _, _}) -> rfc6120;
 get_mod({shim, _}) -> xep0131;
@@ -1442,6 +1468,7 @@ get_mod({vcard_photo, _, _, _}) -> xep0054;
 get_mod({ps_publish, _, _}) -> xep0060;
 get_mod({search_item, _, _, _, _, _}) -> xep0055;
 get_mod({xcaptcha, _}) -> xep0158;
+get_mod({avatar_data, _}) -> xep0084;
 get_mod({roster_query, _, _}) -> rfc6121;
 get_mod({vcard_key, _, _}) -> xep0054;
 get_mod({sm_r, _}) -> xep0198;
@@ -1466,7 +1493,7 @@ get_mod({gone, _}) -> rfc6120;
 get_mod({sasl_response, _}) -> rfc6120;
 get_mod({chatstate, _}) -> xep0085;
 get_mod({muc_unique, _}) -> xep0045;
-get_mod({muc_unsubscribe, _}) -> p1_mucsub;
+get_mod({receipt_response, _}) -> xep0184;
 get_mod({sic, _, _, _}) -> xep0279;
 get_mod({sm_resumed, _, _, _}) -> xep0198;
 get_mod({offline_item, _, _}) -> xep0013;
@@ -1484,12 +1511,14 @@ get_mod({vcard_email, _, _, _, _, _, _}) -> xep0054;
 get_mod({ps_affiliation, _, _, _, _}) -> xep0060;
 get_mod({feature_csi, _}) -> xep0352;
 get_mod({rsm_set, _, _, _, _, _, _, _}) -> xep0059;
+get_mod({avatar_meta, _, _}) -> xep0084;
 get_mod({unblock, _}) -> xep0191;
 get_mod({stat_error, _, _}) -> xep0039;
 get_mod({sasl_abort}) -> rfc6120;
 get_mod({starttls_proceed}) -> rfc6120;
 get_mod({compressed}) -> xep0138;
 get_mod({stream_error, _, _}) -> rfc6120;
+get_mod({receipt_request}) -> xep0184;
 get_mod({upload_request_0, _, _, _, _}) -> xep0363;
 get_mod({privacy_item, _, _, _, _, _, _, _, _}) ->
     xep0016;
@@ -1497,4 +1526,12 @@ get_mod({starttls_failure}) -> rfc6120;
 get_mod({bob_data, _, _, _, _}) -> xep0231;
 get_mod({bind, _, _}) -> rfc6120;
 get_mod({rsm_first, _, _}) -> xep0059;
+get_mod({caps, _, _, _, _}) -> xep0115;
+get_mod({xdata, _, _, _, _, _, _}) -> xep0004;
+get_mod({ps_options, _, _, _, _}) -> xep0060;
+get_mod({muc_admin, _}) -> xep0045;
+get_mod({muc_unsubscribe, _, _}) -> p1_mucsub;
+get_mod({search, _, _, _, _, _, _, _}) -> xep0055;
+get_mod({expire, _, _}) -> xep0023;
+get_mod({privilege, _, _}) -> xep0356;
 get_mod(Record) -> xmpp_codec_external:lookup(Record).
